/*
 *     terraVox: an experimental voxel landscape engine
 *
 *     Copyright (C) 1999/2000 Alex J. Champandard
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *     The tutorial that accompanies this source code can be found at
 *         http://www.flipcode.com
 *
 *     You can also visit the terraVox web site at
 *         http://atlas.cs.york.ac.uk/~ajc116/terraVox
 *
 *     And if you wish to contact me for any reason, I can be reached at
 *         alexjc@altavista.com
 *
 */

#ifndef __MATRIX_H
#define __MATRIX_H

#include "vector.h"

class matrix
{

	vector m[4];

public:

	inline matrix operator*(const matrix &b) const;
	inline matrix operator+(const matrix &b) const;
	inline matrix operator-(const matrix &b) const;
	inline matrix operator-() const;

	inline matrix operator/(const double d) const;
	inline matrix operator*(const double d) const;
	inline vector operator*(const vector &v) const;

	void matrix::transpose();

	      vector &operator[](const int i)       { return m[i]; }
	const vector &operator[](const int i) const { return m[i]; }

	matrix &operator=(const matrix &b) { m = b.m; return *this; }
        void matrix::translate(double x,double y, double z);

	matrix(const matrix &b):m(b.m)  { }
	matrix(	const double a11, const double a12, const double a13,
		const double a21, const double a22, const double a23,
		const double a31, const double a32, const double a33)
	{
		m[0][0]=a11; m[0][1]=a12; m[0][2]=a13; m[0][3] = 0;
		m[1][0]=a21; m[1][1]=a22; m[1][2]=a23; m[1][3] = 0;
		m[2][0]=a31; m[2][1]=a32; m[2][2]=a33; m[2][3] = 0;
		m[3][0]=  0; m[3][1]=  0; m[3][2]=  0; m[3][3] = 1;
	}
	 matrix() {}
	~matrix() {}
};

inline matrix matrix::operator*(const matrix &b) const return r
{
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++) {
     	            r[i][j] = m[i][0] * b[0][j]
                            + m[i][1] * b[1][j]
                            + m[i][2] * b[2][j]
                            + m[i][3] * b[3][j];
                }
}

inline matrix matrix::operator+(const matrix &b) const return r
{
	r[0] = m[0]+b[0];
	r[1] = m[1]+b[1];
	r[2] = m[2]+b[2];
	r[3] = m[3]+b[3];
}
inline matrix matrix::operator-(const matrix &b) const return r
{
	r[0] = m[0]-b[0];
	r[1] = m[1]-b[1];
	r[2] = m[2]-b[2];
	r[3] = m[3]-b[3];
}
inline matrix matrix::operator-() const return r
{
	r[0] = -m[0];
	r[1] = -m[1];
	r[2] = -m[2];
	r[2] = -m[3];
}


// scalar
inline matrix operator*(const double d, const matrix &a) return r
{
	r[0] = a[0]*d;
	r[1] = a[1]*d;
	r[2] = a[2]*d;
	r[3] = a[3]*d;
}
inline matrix matrix::operator/(const double d) const return r
{
        double id = 1/d;
	r[0] = m[0]*id;
	r[1] = m[1]*id;
	r[2] = m[2]*id;
	r[3] = m[2]*id;
}
inline matrix matrix::operator*(const double d) const return r
{
	r[0] = m[0]*d;
	r[1] = m[1]*d;
	r[2] = m[2]*d;
	r[3] = m[3]*d;
}
inline vector matrix::operator*(const vector &v) const return r
{
     r[0] = v[0] * m[0][0] + v[1] * m[1][0] + v[2] * m[2][0] + m[3][0];
     r[1] = v[0] * m[0][1] + v[1] * m[1][1] + v[2] * m[2][1] + m[3][1];
     r[2] = v[0] * m[0][2] + v[1] * m[1][2] + v[2] * m[2][2] + m[3][2];
     r[3] = m[3][3];
     return r;
}

// other
inline matrix transpose(const matrix &b) return r
{
	r[0][0] = b[0][0];
	r[0][1] = b[1][0];
	r[0][2] = b[2][0];

	r[1][0] = b[0][1];
	r[1][1] = b[1][1];
	r[1][2] = b[2][1];

	r[2][0] = b[0][2];
	r[2][1] = b[1][2];
	r[2][2] = b[2][2];
}

// non member

inline double det(const matrix &b)
{
	return
	  b[0][0]*b[1][1]*b[2][2]
	+ b[0][1]*b[1][2]*b[2][0]
	+ b[0][2]*b[1][0]*b[2][1]
	- b[0][2]*b[1][1]*b[2][0]
	- b[0][0]*b[1][2]*b[2][1]
	- b[0][1]*b[1][0]*b[2][2];
}

inline matrix rotX(const double theta)
{
	const double c = cos(theta);
	const double s = sin(theta);
	return matrix(	 1, 0, 0,
			 0, c, s,
			 0,-s, c);
}
inline matrix rotY(const double theta)
{
	const double c = cos(theta);
	const double s = sin(theta);
	return matrix(	 c, 0,-s,
			 0, 1, 0,
			 s, 0, c);
}

inline matrix rotZ(const double theta)
{
	const double c = cos(theta);
	const double s = sin(theta);
	return matrix(	 c, s, 0,
			-s, c, 0,
			 0, 0, 1);
}
inline matrix scale(const vector &s)
{
	return matrix(s[0],   0,  0,
			 0,s[1],  0,
		   	 0,   0,s[2]);
}

inline matrix translate(double x, double y, double z)
{
   matrix t;
   t[0][0] = 1;   t[0][1] = 0;   t[0][2] = 0;   t[0][3] = 0;
   t[1][0] = 0;   t[1][1] = 1;   t[1][2] = 0;   t[1][3] = 0;
   t[2][0] = 0;   t[2][1] = 0;   t[2][2] = 1;   t[2][3] = 0;
   t[3][0] = -x;  t[3][1] = -y;  t[3][2] = -z;  t[3][3] = 1;
   return t;
}

#endif
